@@ -4,28 +4,20 @@ import base64
 import cv2
 import os
 import re
-import numpy as np
+import numpy as npupda
 import concurrent.futures
+import urllib.parse
 
 from tqdm import tqdm
 
 from ximilar.client.constants import *
+from ximilar.client.exceptions import XimilarClientException
+from ximilar.client.utils.decorators import retry_when
 
 CONFIG_ENDPOINT = "account/v2/config/"
 BASE64_HEADER_PATTERN = re.compile(r"^data:image/(\w+);base64,")
 
 
-class XimilarClientException(Exception):
-    def __init__(self, code, msg=None):
-        Exception.__init__(self, msg)
-
-        self.code = code
-        self.msg = msg
-
-    def __str__(self):
-        return self.msg
-
-
 class RestClient(object):
     """
     Parent class that implements HTTP GET, POST, DELETE methods with requests lib and loading images to base64.
@@ -38,7 +30,11 @@ class RestClient(object):
         self.cache = {}
         self.endpoint = endpoint
         self.max_image_size = max_image_size
-        self.headers = {"Content-Type": "application/json", "Authorization": self.get_token_header(self.token)}
+        self.headers = {
+            "Content-Type": "application/json",
+            "Authorization": self.get_token_header(self.token),
+            "User-Agent": "Ximilar Client/Python"
+        }
         self.check_resource(resource_name)
         self.request_timeout = request_timeout
 
@@ -68,6 +64,7 @@ class RestClient(object):
         url = "/".join(map(lambda x: str(x).rstrip("/").lstrip("/"), args))
         return url
 
+    @retry_when(ConnectionError)
     def get(self, api_endpoint, data=None, params=None):
         """
         Call the http GET request with data.
@@ -85,6 +82,7 @@ class RestClient(object):
         )
         return result.json()
 
+    @retry_when(ConnectionError)
     def post(self, api_endpoint, data=None, files=None, params=None, method=requests.post):
         """
         Call the http POST request with data.
@@ -97,13 +95,11 @@ class RestClient(object):
         :return: json response
         """
         self.invalidate()
-
         if data is not None and files is not None:
             raise Exception("Unable to send data along with files with python requests library!")
 
         if data is not None:
             data = json.dumps(data)
-
         result = method(
             self.urljoin(self.endpoint, api_endpoint),
             params=params,
@@ -120,12 +116,14 @@ class RestClient(object):
         except ValueError as e:
             return None
 
+    @retry_when(ConnectionError)
     def put(self, api_endpoint, data=None, files=None, params=None):
         """
         Call the http PUT request with data
         """
         return self.post(api_endpoint, data=data, files=files, params=params, method=requests.put)
 
+    @retry_when(ConnectionError)
     def delete(self, api_endpoint, data=None, params=None):
         """
         Call the http DELETE request with data.
@@ -136,11 +134,12 @@ class RestClient(object):
         """
         self.invalidate()
 
+        url = urllib.parse.urljoin(self.endpoint, api_endpoint)
         result = requests.delete(
-            self.endpoint + api_endpoint, params=params, headers=self.headers, data=data, timeout=self.request_timeout
+            url, params=params, headers=self.headers, data=data, timeout=self.request_timeout
         )
 
-        if result.status_code == HTTP_NO_COTENT_204:
+        if result.status_code == HTTP_NO_CONTENT_204:
             return result
 
         return result.json()
@@ -414,14 +413,14 @@ class RestClient(object):
         return self.get(CONFIG_ENDPOINT + config_type + version)
 
     @staticmethod
-    def download_image(url, destination=""):
+    def download_image(url, destination="", name=None):
         """
         Download image from url to the destination
         :param url: url to the image
         :param destination: where the image will be stored
         :return: None
         """
-        f_name = url.split("/")[-1]
+        f_name = url.split("/")[-1] if name is None else name
         f_dest = os.path.join(destination, f_name)
         f_dest += ".jpg" if "." not in f_name else ""
 
@@ -459,15 +458,34 @@ class RestClient(object):
 
         results = []
         if output:
+            status = {"answer_records": 0, "records": 0, "error": 0, "skipped_records": 0}
             with tqdm(total=len(records)) as pbar:
                 for future in futures:
                     result = future["future"].result()
+                    self.update_status(status, result)
                     results.append(result)
                     pbar.update(future["size"])
+            print(status)
         else:
             results = [future["future"].result() for future in futures]
         return results
 
+    def update_status(self, status, result):
+        if "status" in result:
+            if isinstance(result["status"], int):
+                if result["status"] >= 300:
+                    status["error"] += 1
+            if isinstance(result["status"], dict):
+                if "code" in result["status"]:
+                    if result["status"]["code"] >= 300:
+                        status["error"] += 1
+        if "records" in result:
+            status["records"] += len(result["records"])
+        if "skipped_records" in result:
+            status["skipped_records"] += len(result["skipped_records"])
+        if "answer_records" in result:
+            status["answer_records"] += len(result["answer_records"])
+
     def batch(self, iterable, n=1):
         """
         Divides list to batches of size n.
