@@ -3,7 +3,7 @@ import json
 
 from ximilar.client import RestClient
 from ximilar.client.constants import *
-
+from ximilar.client.exceptions import XimilarClientInvalidDataException
 
 LABEL_ENDPOINT = "recognition/v2/label/"
 TASK_ENDPOINT = "recognition/v2/task/"
@@ -189,7 +189,7 @@ class RecognitionClient(RestClient):
             RESULT_OK,
         )
 
-    def training_images_iter(self, verification=None):
+    def training_images_iter(self, verification=None, batch_size=1):
         """
         Get iterator overall all of images from workspace.
         :param verification: optional, integer which says how many verifications should have the images
@@ -197,12 +197,48 @@ class RecognitionClient(RestClient):
         """
         images, next_page, status = self.get_training_images(verification=verification)
         while images:
-            for image in images:
-                yield image
+            for image in self.batch(images, batch_size):
+                if batch_size == 1:
+                    yield image[0]
+                else:
+                    yield image
             if not next_page:
                 break
             images, next_page, status = self.get_training_images(next_page)
 
+    def modify_images(self, images=None, labels=None, add_labels=None, remove_labels=None):
+        """
+        Modifies images in batch. Enter either list of images OR labels to be applied as a filter.
+        :param labels:
+        Filter images by these labels given as a list of IDs
+        :param images:
+        List of image IDs
+        :param add_labels:
+        List of label IDs to assign
+        :param remove_labels:
+        List of label IDs to remove
+        :return: int
+        Number of processed records
+        """
+        if images is None:
+            images = []
+        if labels is None:
+            labels = []
+        if remove_labels is None:
+            remove_labels = []
+        if add_labels is None:
+            add_labels = []
+
+        if not (labels or images):
+            raise XimilarClientInvalidDataException("Either images or labels must be specified")
+
+        return self.post(api_endpoint=IMAGE_ENDPOINT + "update", data={
+            "images": images,
+            "labels": labels,
+            "labels-add": add_labels,
+            "labels-remove": remove_labels,
+        })
+
     def get_labels_by_substring(self, name):
         """
         Return all labels of the workspace which contains the substring.
@@ -282,36 +318,7 @@ class RecognitionClient(RestClient):
             metadata = record[META_DATA] if META_DATA in record and record[META_DATA] else {}
             test_image = record[TEST_IMAGE] if TEST_IMAGE in record else False
 
-            if IMG_DATA in record:
-                data = {
-                    "base64": self.cv2img_to_base64(record[IMG_DATA], record[COLOR_SPACE], resize=not noresize),
-                    NORESIZE: noresize_on_server,
-                    TEST_IMAGE: test_image,
-                    META_DATA: metadata,
-                }
-            elif FILE in record:
-                # We cannot send files to request along with json data (for workspace)
-                # That is why we load image from disk to base64 representation
-                data = {
-                    "base64": self.load_base64_file(record[FILE], resize=not noresize),
-                    NORESIZE: noresize_on_server,
-                    TEST_IMAGE: test_image,
-                    META_DATA: metadata,
-                }
-            elif BASE64 in record:
-                data = {
-                    "base64": record[BASE64].decode("utf-8"),
-                    NORESIZE: noresize_on_server,
-                    TEST_IMAGE: test_image,
-                    META_DATA: metadata,
-                }
-            elif URL in record:
-                data = {
-                    "base64": self.load_url_image(record[URL], resize=not noresize),
-                    NORESIZE: noresize_on_server,
-                    TEST_IMAGE: test_image,
-                    META_DATA: metadata,
-                }
+            data = self._create_image_data(record, noresize, noresize_on_server, test_image, metadata)
 
             image_json = self.post(IMAGE_ENDPOINT, files=files, data=data)
 
@@ -331,6 +338,43 @@ class RecognitionClient(RestClient):
             images.append(image)
         return images, worst_status
 
+    def _create_image_data(self, record, noresize, noresize_on_server, test_image, metadata):
+        if IMG_DATA in record:
+            return {
+                "base64": self.cv2img_to_base64(record[IMG_DATA], record[COLOR_SPACE], resize=not noresize),
+                NORESIZE: noresize_on_server,
+                TEST_IMAGE: test_image,
+                META_DATA: metadata,
+            }
+
+        if FILE in record:
+            # We cannot send files to request along with json data (for workspace)
+            # That is why we load image from disk to base64 representation
+            return {
+                "base64": self.load_base64_file(record[FILE], resize=not noresize),
+                NORESIZE: noresize_on_server,
+                TEST_IMAGE: test_image,
+                META_DATA: metadata,
+            }
+
+        if BASE64 in record:
+            return {
+                "base64": record[BASE64].decode("utf-8"),
+                NORESIZE: noresize_on_server,
+                TEST_IMAGE: test_image,
+                META_DATA: metadata,
+            }
+
+        if URL in record:
+            return {
+                "base64": self.load_url_image(record[URL], resize=not noresize),
+                NORESIZE: noresize_on_server,
+                TEST_IMAGE: test_image,
+                META_DATA: metadata,
+            }
+
+        raise Exception("No _file, _url, _base64, _img_data in record!")
+
     def construct_data(self, records=[], task_id=None, version=None, store_images=None):
         if len(records) == 0:
             raise Exception("Please specify at least one record in classify method!")
@@ -552,7 +596,6 @@ class Label(RecognitionClient):
 
         label_json = self.get(LABEL_ENDPOINT + self.id)
         self.images_count = label_json[IMAGES_COUNT]
-
         return self.images_count
 
     # TODO: get object count
@@ -565,18 +608,10 @@ class Label(RecognitionClient):
         :param page_url: optional, select the specific page of images, default first page
         :return: (list of images, next_page)
         """
-        url = (
-            page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")
-            if page_url
-            else IMAGE_ENDPOINT + "?label=" + self.id
-        )
-        result = self.get(url)
+        if page_url is None:
+            page_url = IMAGE_ENDPOINT + "?label=" + self.id
 
-        return (
-            [Image(self.token, self.endpoint, image_json) for image_json in result[RESULTS]],
-            result[NEXT],
-            RESULT_OK,
-        )
+        return super().get_training_images(page_url=page_url, verification=verification)
 
     def upload_images(self, records):
         """
@@ -657,6 +692,7 @@ class Image(RecognitionClient):
         )
         # file path after calling download_image on this object
         self._file = None
+        self._objects = []
 
     def __str__(self):
         return self.thumb_img_path
@@ -695,13 +731,13 @@ class Image(RecognitionClient):
         """
         return self.post(IMAGE_ENDPOINT + self.id + "/remove-label/", data={LABEL_ID: label_id})
 
-    def download(self, destination=""):
+    def download(self, destination="", name=None):
         """
         Download image to the destination and store path to the _file of the object.
         :param destination: path on the disk
         :return: path of the file locally
         """
-        self._file = super().download_image(self.img_path, destination=destination)
+        self._file = super().download_image(self.img_path, destination=destination, name=name)
         return self._file
 
     def _ensure_meta_data(self):
